from typing import Callable, Any
from pymongo.read_concern import ReadConcern
from pymongo.write_concern import WriteConcern
from pymongo.read_preferences import ReadPreference

from navalmartin_mir_db_utils.dbs.mongodb_session import MongoDBSession
from navalmartin_mir_db_utils.dbs.dbs_utils import DB_ERROR, DB_INFO


async def run_transaction(mdb_session: MongoDBSession,
                          async_callback: Callable,
                          callback_args: dict,
                          max_commit_time_ms: int,
                          read_concern: ReadConcern, write_concern: WriteConcern,
                          read_preference: ReadPreference,
                          with_log: bool = False,
                          transaction_result_handler: Callable = None) -> Any:
    """
    mdb_session: The MongoDB session
    async_callback: The function callback to be used
    callback_args: Dictionary with arguments to pass in the callback
    max_commit_time_ms: The maximum commit time for the transaction
    read_concern: The ReadConcern
    write_concern: The  WriteConcern
    read_preference: The ReadPreference
    with_log: If True it prints information massages on std output
    transaction_result_handler: Callable that specifies how to handle the transaction result

    Returns
    -------

    The transaction result
    """
    try:

        if with_log:
            print(f"{DB_INFO} Starting transaction")

        # Start a client session.
        async with await mdb_session.client.start_session() as session:

            # Use with_transaction to start a transaction,
            # execute the callback, and commit (or abort on error).
            transaction_result = await session.with_transaction(
                lambda s: async_callback(session, callback_args),
                read_concern=read_concern,
                write_concern=write_concern,
                read_preference=read_preference,
                max_commit_time_ms=max_commit_time_ms
            )

            if transaction_result_handler is not None:
                transaction_result = await transaction_result_handler(transaction_result)

        if with_log:
            print(f"{DB_INFO} End transaction")

        return transaction_result
    except Exception as e:
        if with_log:
            print(f"{DB_ERROR} Session with transaction raised an exception. Exception message={str(e)}")
        raise e



from typing import Any, List


class InvalidObjectIdException(Exception):
    def __init__(self, oid: str):
        self.message = f"The given id {oid} does not represent a valid bson.ObjectId"

    def __str__(self) -> str:
        return self.message


class ResourceNotFoundException(Exception):
    def __init__(self, search_criteria: str):
        self.message = f"Resource with search_criteria {search_criteria} not found"

    def __str__(self) -> str:
        return self.message


class ResourceNotUpdatedException(Exception):
    def __init__(self, resource_id: str):
        self.message = f"Resource with id {resource_id} not updated"

    def __str__(self) -> str:
        return self.message


class ResourceExistsException(Exception):
    def __init__(self, resource_id: str):
        self.message = f"Resource with id {resource_id} exists"

    def __str__(self) -> str:
        return self.message


class DBInsertFailedException(Exception):
    def __init__(self, collection_name: str):
        self.message = f"Insert operation failed for collection {collection_name}"

    def __str__(self) -> str:
        return self.message


class DBUpdateFailedException(Exception):
    def __init__(self, collection_name: str):
        self.message = f"Update operation failed for collection {collection_name}"

    def __str__(self) -> str:
        return self.message


class InvalidAttributeValueException(Exception):
    def __init__(self, attribute_name: str, attribute_value: Any, *, allowed_values: List = None,
                 range: tuple = None):
        if allowed_values is not None:
            self.message = f"Attribute {attribute_name} has value {attribute_value} not in {allowed_values}"
        elif range is not None:
            self.message = f"Attribute {attribute_name} has value {attribute_value} not in range [{range[0]}, {range[1]}]"
        else:
            self.message = f"Attribute {attribute_name} has value {attribute_value} which is not valid"
        self.attribute_name = attribute_name
        self.attribute_value = attribute_value

    def __str__(self):
        return self.message


class InvalidMongoDBOperatorException(Exception):
    def __init__(self, operator: str, extra_message: str = ""):
        self.message = f"The operator {operator} is an invalid MongoDB operator"

        if extra_message != "" or extra_message is not None:
            self.message += " " + extra_message

    def __str__(self):
        return self.message




from .data_views.indexed_item_data_view_base import IndexedItemDataViewBase
from .data_views.user_data_view_base import UserDataViewBase
from .task_performance_result_schema import TaskPerformanceResultSchema
from .task_result_schema import TaskResultSchema, ErrorType
from .task_configuration_schema import TaskConfigurationSchema





import asyncio
import bson
from navalmartin_mir_db_utils.dbs.mongodb_session import MongoDBSession
from navalmartin_mir_db_utils.crud.mongodb_crud_ops import ReadEntityCRUDAPI
from navalmartin_mir_db_utils.utils.exceptions import ResourceNotFoundException
from navalmartin_mir_db_utils.crud.crud_utils import get_one_result_or_raise

COLLECTION_NAME = "YOUR_COLLECTION_NAME"
MONGODB_URL = "YOUR_MONGODB_URL"
MONGO_DB_NAME_FROM = "YOUR_MONGODB_NAME"


async def query_db(mongodb_session: MongoDBSession, criteria: dict,
                   projection: dict,
                   collection_name: str):
    query_result = ReadEntityCRUDAPI.find(criteria=criteria, projection=projection,
                                          db_session=mongodb_session,
                                          collection_name=collection_name)
    docs = [doc async for doc in query_result]
    return docs


async def count_docs(mongodb_session: MongoDBSession, criteria: dict,
                     collection_name: str):
    query_result = await ReadEntityCRUDAPI.count_documents(criteria=criteria,
                                                           db_session=mongodb_session,
                                                           collection_name=collection_name)
    return query_result


async def query_db_or_raise(mongodb_session: MongoDBSession, criteria: dict,
                            projection: dict,
                            collection_name: str):
    query_result = await get_one_result_or_raise(crud_handler=ReadEntityCRUDAPI(collection_name=collection_name),
                                                 projection=projection,
                                                 criteria=criteria,
                                                 db_session=mongodb_session)

    return query_result


async def run_examples(mir_db_session_from: MongoDBSession):
    result = await query_db(mongodb_session=mir_db_session_from,
                            criteria={'_id': bson.ObjectId('63ebc9f94c092a48bd179ae7')},
                            projection={},
                            collection_name=COLLECTION_NAME)
    print(result)
    n_docs = await count_docs(mongodb_session=mir_db_session_from,
                              criteria={},
                              collection_name=COLLECTION_NAME)
    print(n_docs)

    try:
        result = await query_db_or_raise(mongodb_session=mir_db_session_from,
                                         criteria={'survey_idx': bson.ObjectId('63ad64252c853ee163fc6a63')},
                                         projection={'original_filename': 1},
                                         collection_name=COLLECTION_NAME)
    except ResourceNotFoundException as e:
        print(str(e))


def main():
    mir_db_session_from = MongoDBSession(mongodb_url=MONGODB_URL,
                                         db_name=MONGO_DB_NAME_FROM)

    asyncio.run(run_examples(mir_db_session_from=mir_db_session_from))

    
if __name__ == '__main__':
    main()



from .mongodb_crud_ops import UpdateEntityCRUDAPI, ReadEntityCRUDAPI, DeleteEntityCRUDAPI, CreateEntityCRUDAPI
from .crud_utils import insert_one_or_fail, if_resource_found_raise, get_one_result_or_raise, update_one_or_raise
from .mongodb_response_adaptors import QueryResponseAdaptor
from .task_utils import write_to_task_error_collection



from pydantic import BaseModel
from typing import List, Any


class QueryResponseAdaptor:
    @staticmethod
    def adapt_one_model_as_dict(model: BaseModel, exclude_vals: List[Any]) -> dict:

        result = {}

        for item in model.__fields_set__:
            if model.__dict__[item] in exclude_vals:
                continue
            result[item] = model.__dict__[item]
        return result



from typing import Any, List
from navalmartin_mir_db_utils.crud.mongodb_crud_ops import CreateEntityCRUDAPI
from navalmartin_mir_db_utils.dbs.mongodb_session import MongoDBSession


async def write_to_task_error_collection(event: dict,
                                         error_msg: str,
                                         task_id: str,
                                         error_type: str,
                                         lambda_arn: str,
                                         lambda_request_id: str,
                                         db_session: MongoDBSession,
                                         collection_name: str) -> Any:
    response = await CreateEntityCRUDAPI.insert_one(data={"event": event,
                                                          "task_id": task_id,
                                                          'error_msg': error_msg,
                                                          "error_type": error_type,
                                                          "lambda_arn": lambda_arn,
                                                          "lambda_request_id": lambda_request_id},
                                                    db_session=db_session,
                                                    collection_name=collection_name)
    return response



"""module task_performance_result_schema. Models basic schema
for monitoring the performance of a task. The schema is meant
to be embedded in a task document specified by the client code

"""

import datetime
from pydantic import BaseModel, Field


class TaskPerformanceResultSchema(BaseModel):

    latency: float = Field(title='latency',
                           description="The overall time the task lasted",
                           default=0.0)
    disk_util: dict = Field(title='disk_util',
                            description="The disk utilization of the task",
                            default={})
    cpu_util: float = Field(title='cpu_util',
                            description="The percentage of the CPU utilization",
                            default=0.0)
    gpu_util: float = Field(title='gpu_util',
                            description="The percentage of the GPU utilization",
                            default=0.0)
    cpu_mem_util: float = Field(title='cpu_mem_util',
                                description="The percentage of the CPU-RAM utilization",
                                default=0.0)
    gpu_mem_util: float = Field(title='gpu_mem_util',
                                description="The percentage of the GPU-RAM utilization",
                                default=0.0)
    started_at: str = Field(title='started_at', description="Date/time the task started UTC format",
                            default=datetime.datetime.utcnow())
    ended_at: str = Field(title='ended_at',
                          description="Date/time the task ended UTC format",
                          default=datetime.datetime.utcnow())
